@@ -21,12 +21,17 @@
 #include <netdb.h>
 #include <signal.h>
 #include <sys/time.h>
+#include <sys/wait.h>
 #include <netinet/in.h>
 #include <glib.h>
 
 #include "cache.h"
 #include "opts.h"
 
+#if FUSE_VERSION >= 23
+#define SSHFS_USE_INIT
+#endif
+
 #define SSH_FXP_INIT                1
 #define SSH_FXP_VERSION             2
 #define SSH_FXP_OPEN                3
@@ -84,12 +89,30 @@
 
 #define MAX_REPLY_LEN (1 << 17)
 
+#define RENAME_TEMP_CHARS 8
+
+#define SFTP_SERVER_PATH "/usr/lib/sftp-server"
+
+static unsigned int randseed;
+
 static int infd;
 static int outfd;
+static int connver;
+static int server_version;
+static unsigned ssh_ver = 2;
+static char *sftp_server = NULL;
 static int debug = 0;
+static int reconnect = 0;
 static int sync_write = 0;
 static int sync_read = 0;
+static int rename_workaround = 0;
+static int detect_uid = 0;
+static unsigned remote_uid;
+static unsigned local_uid;
+static int remote_uid_detected = 0;
 static char *base_path;
+static char *host;
+static unsigned blksize = 4096;
 
 struct buffer {
     uint8_t *p;
@@ -110,6 +133,7 @@ struct request {
     sem_t ready;
     uint8_t reply_type;
     int replied;
+    int error;
     struct buffer reply;
     struct timeval start;
     void *data;
@@ -134,12 +158,14 @@ struct sshfs_file {
     struct read_chunk *readahead;
     off_t next_pos;
     int is_seq;
+    int connver;
 };
 
 static GHashTable *reqtab;
 static pthread_mutex_t lock;
 static int processing_thread_started;
 
+
 static struct opt ssh_opts[] = {
     { .optname = "AddressFamily"                    },
     { .optname = "BatchMode"                        },
@@ -151,7 +177,7 @@ static struct opt ssh_opts[] = {
     { .optname = "Compression"                      },
     { .optname = "CompressionLevel"                 },
     { .optname = "ConnectionAttempts"               },
-    { .optname = "ConnectionTimeout"                },
+    { .optname = "ConnectTimeout"                   },
     { .optname = "GlobalKnownHostsFile"             },
     { .optname = "GSSAPIAuthentication"             },
     { .optname = "GSSAPIDelegateCredentials"        },
@@ -168,7 +194,6 @@ static struct opt ssh_opts[] = {
     { .optname = "PasswordAuthentication"           },
     { .optname = "Port"                             },
     { .optname = "PreferredAuthentications"         },
-    { .optname = "Protocol"                         },
     { .optname = "ProxyCommand"                     },
     { .optname = "PubkeyAuthentication"             },
     { .optname = "RhostsRSAAuthentication"          },
@@ -191,6 +216,11 @@ enum {
     SOPT_SYNC_READ,
     SOPT_MAX_READ,
     SOPT_DEBUG,
+    SOPT_RECONNECT,
+    SOPT_WORKAROUND,
+    SOPT_IDMAP,
+    SOPT_PROTOCOL,
+    SOPT_SFTPSERVER,
     SOPT_LAST /* Last entry in this list! */
 };
 
@@ -201,6 +231,11 @@ static struct opt sshfs_opts[] = {
     [SOPT_SYNC_READ]  = { .optname = "no_readahead" },
     [SOPT_MAX_READ]   = { .optname = "max_read" },
     [SOPT_DEBUG]      = { .optname = "sshfs_debug" },
+    [SOPT_RECONNECT]  = { .optname = "reconnect" },
+    [SOPT_WORKAROUND] = { .optname = "workaround" },
+    [SOPT_IDMAP]      = { .optname = "idmap" },
+    [SOPT_PROTOCOL]   = { .optname = "ssh_protocol" },
+    [SOPT_SFTPSERVER] = { .optname = "sftp_server" },
     [SOPT_LAST]       = { .optname = NULL }
 };
 
@@ -442,7 +477,7 @@ static inline int buf_get_string(struct buffer *buf, char **str)
     return 0;
 }
 
-static int buf_get_attrs(struct buffer *buf, struct stat *stbuf)
+static int buf_get_attrs(struct buffer *buf, struct stat *stbuf, int *flagsp)
 {
     uint32_t flags;
     uint64_t size = 0;
@@ -454,6 +489,8 @@ static int buf_get_attrs(struct buffer *buf, struct stat *stbuf)
 
     if (buf_get_uint32(buf, &flags) == -1)
         return -1;
+    if (flagsp)
+        *flagsp = flags;
     if ((flags & SSH_FILEXFER_ATTR_SIZE) &&
         buf_get_uint64(buf, &size) == -1)
         return -1;
@@ -484,10 +521,18 @@ static int buf_get_attrs(struct buffer *buf, struct stat *stbuf)
             buf_free(&tmp);
         }
     }
+
+    if (remote_uid_detected && uid == remote_uid)
+        uid = local_uid;
+
     memset(stbuf, 0, sizeof(struct stat));
     stbuf->st_mode = mode;
     stbuf->st_nlink = 1;
     stbuf->st_size = size;
+    if (blksize) {
+        stbuf->st_blksize = blksize;
+        stbuf->st_blocks = ((size + blksize - 1) & ~(blksize - 1)) >> 9;
+    }
     stbuf->st_uid = uid;
     stbuf->st_gid = gid;
     stbuf->st_atime = atime;
@@ -513,7 +558,7 @@ static int buf_get_entries(struct buffer *buf, fuse_cache_dirh_t h,
             return -1;
         if (buf_get_string(buf, &longname) != -1) {
             free(longname);
-            if (buf_get_attrs(buf, &stbuf) != -1) {
+            if (buf_get_attrs(buf, &stbuf, NULL) != -1) {
                 filler(h, name, &stbuf);
                 err = 0;
             }
@@ -544,9 +589,14 @@ static int start_ssh(char *host)
         perror("failed to fork");
         return -1;
     } else if (pid == 0) {
+        int devnull;
         int argctr = 0;
         char *ssh_args[sizeof(ssh_opts)/sizeof(struct opt) + 32];
         char *ssh_cmd;
+        char veropt[64];
+
+        devnull = open("/dev/null", O_WRONLY);
+
         if (sshfs_opts[SOPT_SSHCMD].present)
             ssh_cmd = sshfs_opts[SOPT_SSHCMD].value;
         else
@@ -554,15 +604,31 @@ static int start_ssh(char *host)
 
         if (dup2(outpipe[0], 0) == -1 || dup2(inpipe[1], 1) == -1) {
             perror("failed to redirect input/output");
-            exit(1);
+            _exit(1);
         }
+        if (!debug && devnull != -1)
+            dup2(devnull, 2);
+
+        close(devnull);
         close(inpipe[0]);
         close(inpipe[1]);
         close(outpipe[0]);
         close(outpipe[1]);
 
+        switch (fork()) {
+        case -1:
+            perror("failed to fork");
+            _exit(1);
+        case 0:
+            break;
+        default:
+            _exit(0);
+        }
+        chdir("/");
+
         ssh_args[argctr++] = ssh_cmd;
-        ssh_args[argctr++] = "-2";
+        sprintf(veropt, "-%i", ssh_ver);
+        ssh_args[argctr++] = veropt;
         ssh_args[argctr++] = "-x";
         ssh_args[argctr++] = "-a";
         ssh_args[argctr++] = "-oClearAllForwardings=yes";
@@ -575,13 +641,21 @@ static int start_ssh(char *host)
             }
         }
         ssh_args[argctr++] = host;
-        ssh_args[argctr++] = "-s";
-        ssh_args[argctr++] = "sftp";
+        if (!sftp_server) {
+            if (ssh_ver == 1)
+                sftp_server = SFTP_SERVER_PATH;
+            else
+                sftp_server = "sftp";
+        }
+        if (ssh_ver != 1 && strchr(sftp_server, '/') == NULL)
+            ssh_args[argctr++] = "-s";
+        ssh_args[argctr++] = sftp_server;
         ssh_args[argctr++] = NULL;
 
         execvp(ssh_cmd, ssh_args);
-        exit(1);
+        _exit(1);
     }
+    waitpid(pid, NULL, 0);
     close(inpipe[1]);
     close(outpipe[0]);
     return 0;
@@ -653,11 +727,9 @@ static int sftp_send(uint8_t type, struct buffer *buf)
     buf_init(&buf2, 5);
     buf_add_uint32(&buf2, buf->len + 1);
     buf_add_uint8(&buf2, type);
-    pthread_mutex_lock(&lock);
     res = do_write(&buf2);
     if (res != -1)
         res = do_write(buf);
-    pthread_mutex_unlock(&lock);
     buf_free(&buf2);
     return res;
 }
@@ -673,7 +745,7 @@ static int do_read(struct buffer *buf)
             perror("read");
             return -1;
         } else if (res == 0) {
-            fprintf(stderr, "end of file read\n");
+            fprintf(stderr, "remote host has disconnected\n");
             return -1;
         }
         size -= res;
@@ -690,12 +762,14 @@ static int sftp_read(uint8_t *type, struct buffer *buf)
     buf_init(&buf2, 5);
     res = do_read(&buf2);
     if (res != -1) {
-        buf_get_uint32(&buf2, &len);
+        if (buf_get_uint32(&buf2, &len) == -1)
+            return -1;
         if (len > MAX_REPLY_LEN) {
             fprintf(stderr, "reply len too large: %u\n", len);
             return -1;
         }
-        buf_get_uint8(&buf2, type);
+        if (buf_get_uint8(&buf2, type) == -1)
+            return -1;
         buf_init(buf, len - 1);
         res = do_read(buf);
     }
@@ -733,12 +807,27 @@ static void chunk_put_locked(struct read_chunk *chunk)
     pthread_mutex_unlock(&lock);
 }
 
-static void *process_requests(void *_data)
+static int clean_req(void *key_, struct request *req)
 {
-    (void) _data;
+    (void) key_;
+
+    req->error = -EIO;
+    if (req->want_reply)
+        sem_post(&req->ready);
+    else {
+        if (req->end_func)
+            req->end_func(req);
+        request_free(req);
+    }
+    return TRUE;
+}
+
+static void *process_requests(void *data_)
+{
+    int res;
+    (void) data_;
 
     while (1) {
-        int res;
         struct buffer buf;
         uint8_t type;
         struct request *req;
@@ -772,17 +861,132 @@ static void *process_requests(void *_data)
             if (req->want_reply)
                 sem_post(&req->ready);
             else {
-                if (req->end_func)
+                if (req->end_func) {
+                    pthread_mutex_lock(&lock);
                     req->end_func(req);
+                    pthread_mutex_unlock(&lock);
+                }
                 request_free(req);
             }
         } else
             buf_free(&buf);
     }
-    kill(getpid(), SIGTERM);
+    if (!reconnect) {
+        /* harakiri */
+        kill(getpid(), SIGTERM);
+    } else {
+        pthread_mutex_lock(&lock);
+        processing_thread_started = 0;
+        close(infd);
+        infd = -1;
+        close(outfd);
+        outfd = -1;
+        g_hash_table_foreach_remove(reqtab, (GHRFunc) clean_req, NULL);
+        connver ++;
+        pthread_mutex_unlock(&lock);
+    }
     return NULL;
 }
 
+static int sftp_init()
+{
+    int res = -1;
+    uint8_t type;
+    uint32_t version;
+    struct buffer buf;
+    buf_init(&buf, 4);
+    buf_add_uint32(&buf, PROTO_VERSION);
+    if (sftp_send(SSH_FXP_INIT, &buf) == -1)
+        goto out;
+    buf_clear(&buf);
+    if (sftp_read(&type, &buf) == -1)
+        goto out;
+    if (type != SSH_FXP_VERSION) {
+        fprintf(stderr, "protocol error\n");
+        goto out;
+    }
+    if (buf_get_uint32(&buf, &version) == -1)
+        goto out;
+
+    server_version = version;
+    DEBUG("Server version: %i\n", server_version);
+    if (version > PROTO_VERSION)
+        fprintf(stderr, "Warning: server uses version: %i, we support: %i\n",
+                version, PROTO_VERSION);
+    res = 0;
+
+ out:
+    buf_free(&buf);
+    return res;
+}
+
+static void sftp_detect_uid()
+{
+    int flags;
+    int id = sftp_get_id();
+    int replid;
+    uint8_t type;
+    struct buffer buf;
+    struct stat stbuf;
+
+    buf_init(&buf, 9);
+    buf_add_uint32(&buf, id);
+    buf_add_string(&buf, ".");
+    if (sftp_send(SSH_FXP_STAT, &buf) == -1)
+        goto out;
+    buf_clear(&buf);
+    if (sftp_read(&type, &buf) == -1)
+        goto out;
+    if (type != SSH_FXP_ATTRS && type != SSH_FXP_STATUS) {
+        fprintf(stderr, "protocol error\n");
+        goto out;
+    }
+    if (buf_get_uint32(&buf, &replid) == -1)
+        goto out;
+    if (replid != id) {
+        fprintf(stderr, "bad reply ID\n");
+        goto out;
+    }
+    if (type == SSH_FXP_STATUS) {
+        uint32_t serr;
+        if (buf_get_uint32(&buf, &serr) == -1)
+            goto out;
+
+        fprintf(stderr, "failed to stat home directory (%i)\n", serr);
+        goto out;
+    }
+    if (buf_get_attrs(&buf, &stbuf, &flags) == -1)
+        goto out;
+
+    if (!(flags & SSH_FILEXFER_ATTR_UIDGID))
+        goto out;
+
+    remote_uid = stbuf.st_uid;
+    local_uid = getuid();
+    remote_uid_detected = 1;
+    DEBUG("remote_uid = %i\n", remote_uid);
+
+ out:
+    if (!remote_uid_detected)
+        fprintf(stderr, "failed to detect remote user ID\n");
+
+    buf_free(&buf);
+}
+
+static int connect_remote(void)
+{
+    int err;
+
+    if (sshfs_opts[SOPT_DIRECTPORT].present)
+        err = connect_to(host, sshfs_opts[SOPT_DIRECTPORT].value);
+    else
+        err = start_ssh(host);
+    if (!err)
+        err = sftp_init();
+
+    return err;
+}
+
 static int start_processing_thread(void)
 {
     int err;
@@ -790,16 +994,33 @@ static int start_processing_thread(void)
     if (processing_thread_started)
         return 0;
 
+    if (outfd == -1) {
+        err = connect_remote();
+        if (err)
+            return -EIO;
+    }
+
     err = pthread_create(&thread_id, NULL, process_requests, NULL);
     if (err) {
         fprintf(stderr, "failed to create thread: %s\n", strerror(err));
-        return -EPERM;
+        return -EIO;
     }
     pthread_detach(thread_id);
     processing_thread_started = 1;
     return 0;
 }
 
+#ifdef SSHFS_USE_INIT
+static void *sshfs_init(void)
+{
+    if (detect_uid)
+        sftp_detect_uid();
+
+    start_processing_thread();
+    return NULL;
+}
+#endif
+
 static int sftp_request_common(uint8_t type, const struct buffer *buf,
                                uint8_t expect_type, struct buffer *outbuf,
                                request_func begin_func, request_func end_func,
@@ -807,42 +1028,48 @@ static int sftp_request_common(uint8_t type, const struct buffer *buf,
 {
     int err;
     struct buffer buf2;
-    uint32_t id = sftp_get_id();
+    uint32_t id;
     struct request *req = g_new0(struct request, 1);
 
-    buf_init(&buf2, buf->len + 4);
-    buf_add_uint32(&buf2, id);
-    buf_add_mem(&buf2, buf->p, buf->len);
-
     req->want_reply = expect_type != 0 ? 1 : 0;
     req->end_func = end_func;
     req->data = data;
     sem_init(&req->ready, 0, 0);
     buf_init(&req->reply, 0);
+    buf_init(&buf2, buf->len + 4);
+    pthread_mutex_lock(&lock);
     if (begin_func)
         begin_func(req);
-    pthread_mutex_lock(&lock);
+    id = sftp_get_id();
+    buf_add_uint32(&buf2, id);
+    buf_add_mem(&buf2, buf->p, buf->len);
     err = start_processing_thread();
+    if (err) {
+        pthread_mutex_unlock(&lock);
+        goto out;
+    }
     g_hash_table_insert(reqtab, (gpointer) id, req);
     gettimeofday(&req->start, NULL);
     DEBUG("[%05i] %s\n", id, type_name(type));
-    pthread_mutex_unlock(&lock);
-    if (err)
-        goto out;
 
     err = -EIO;
     if (sftp_send(type, &buf2) == -1) {
-        pthread_mutex_lock(&lock);
         g_hash_table_remove(reqtab, (gpointer) id);
         pthread_mutex_unlock(&lock);
         goto out;
     }
+    pthread_mutex_unlock(&lock);
+
     if (expect_type == 0) {
         buf_free(&buf2);
         return 0;
     }
 
     sem_wait(&req->ready);
+    if (req->error) {
+        err = req->error;
+        goto out;
+    }
     err = -EPROTO;
     if (req->reply_type != expect_type && req->reply_type != SSH_FXP_STATUS) {
         fprintf(stderr, "protocol error\n");
@@ -881,8 +1108,11 @@ static int sftp_request_common(uint8_t type, const struct buffer *buf,
     }
 
  out:
-    if (end_func)
+    if (end_func) {
+        pthread_mutex_lock(&lock);
         end_func(req);
+        pthread_mutex_unlock(&lock);
+    }
     buf_free(&buf2);
     request_free(req);
     return err;
@@ -911,7 +1141,7 @@ static int sshfs_getattr(const char *path, struct stat *stbuf)
     buf_add_path(&buf, path);
     err = sftp_request(SSH_FXP_LSTAT, &buf, SSH_FXP_ATTRS, &outbuf);
     if (!err) {
-        if (buf_get_attrs(&outbuf, stbuf) == -1)
+        if (buf_get_attrs(&outbuf, stbuf, NULL) == -1)
             err = -EPROTO;
         buf_free(&outbuf);
     }
@@ -924,6 +1154,10 @@ static int sshfs_readlink(const char *path, char *linkbuf, size_t size)
     int err;
     struct buffer buf;
     struct buffer name;
+
+    if (server_version < 3)
+        return -EPERM;
+
     buf_init(&buf, 0);
     buf_add_path(&buf, path);
     err = sftp_request(SSH_FXP_READLINK, &buf, SSH_FXP_NAME, &name);
@@ -1022,6 +1256,10 @@ static int sshfs_symlink(const char *from, const char *to)
 {
     int err;
     struct buffer buf;
+
+    if (server_version < 3)
+        return -EPERM;
+
     /* openssh sftp server doesn't follow standard: link target and
        link name are mixed up, so we must also be non-standard :( */
     buf_init(&buf, 0);
@@ -1054,7 +1292,7 @@ static int sshfs_rmdir(const char *path)
     return err;
 }
 
-static int sshfs_rename(const char *from, const char *to)
+static int sshfs_do_rename(const char *from, const char *to)
 {
     int err;
     struct buffer buf;
@@ -1066,6 +1304,38 @@ static int sshfs_rename(const char *from, const char *to)
     return err;
 }
 
+static void random_string(char *str, int length)
+{
+    int i;
+    for (i = 0; i < length; i++)
+        *str++ = (char)('0' + rand_r(&randseed) % 10);
+    *str = '\0';
+}
+
+static int sshfs_rename(const char *from, const char *to)
+{
+    int err;
+    err = sshfs_do_rename(from, to);
+    if (err == -EPERM && rename_workaround) {
+        size_t tolen = strlen(to);
+        if (tolen + RENAME_TEMP_CHARS < PATH_MAX) {
+            int tmperr;
+            char totmp[PATH_MAX];
+            strcpy(totmp, to);
+            random_string(totmp + tolen, RENAME_TEMP_CHARS);
+            tmperr = sshfs_do_rename(to, totmp);
+            if (!tmperr) {
+                err = sshfs_do_rename(from, to);
+                if (!err)
+                    err = sshfs_unlink(totmp);
+                else
+                    sshfs_do_rename(totmp, to);
+            }
+        }
+    }
+    return err;
+}
+
 static int sshfs_chmod(const char *path, mode_t mode)
 {
     int err;
@@ -1099,9 +1369,26 @@ static int sshfs_truncate(const char *path, off_t size)
     struct buffer buf;
     buf_init(&buf, 0);
     buf_add_path(&buf, path);
-    buf_add_uint32(&buf, SSH_FILEXFER_ATTR_SIZE);
-    buf_add_uint64(&buf, size);
-    err = sftp_request(SSH_FXP_SETSTAT, &buf, SSH_FXP_STATUS, NULL);
+    if (size == 0) {
+        /* If size is zero, use open(..., O_TRUNC), to work around
+           broken sftp servers */
+        struct buffer handle;
+        buf_add_uint32(&buf, SSH_FXF_WRITE | SSH_FXF_TRUNC);
+        buf_add_uint32(&buf, 0);
+        err = sftp_request(SSH_FXP_OPEN, &buf, SSH_FXP_HANDLE, &handle);
+        if (!err) {
+            int err2;
+            buf_finish(&handle);
+            err2 = sftp_request(SSH_FXP_CLOSE, &handle, 0, NULL);
+            if (!err)
+                err = err2;
+            buf_free(&handle);
+        }
+    } else {
+        buf_add_uint32(&buf, SSH_FILEXFER_ATTR_SIZE);
+        buf_add_uint64(&buf, size);
+        err = sftp_request(SSH_FXP_SETSTAT, &buf, SSH_FXP_STATUS, NULL);
+    }
     buf_free(&buf);
     return err;
 }
@@ -1120,7 +1407,13 @@ static int sshfs_utime(const char *path, struct utimbuf *ubuf)
     return err;
 }
 
-static int sshfs_open(const char *path, struct fuse_file_info *fi)
+static inline int sshfs_file_is_conn(struct sshfs_file *sf)
+{
+    return sf->connver == connver;
+}
+
+static int sshfs_open_common(const char *path, mode_t mode,
+                             struct fuse_file_info *fi)
 {
     int err;
     struct buffer buf;
@@ -1135,16 +1428,27 @@ static int sshfs_open(const char *path, struct fuse_file_info *fi)
     else
         return -EINVAL;
 
+    if (fi->flags & O_CREAT)
+        pflags |= SSH_FXF_CREAT;
+
+    if (fi->flags & O_EXCL)
+        pflags |= SSH_FXF_EXCL;
+
+    if (fi->flags & O_TRUNC)
+        pflags |= SSH_FXF_TRUNC;
+
     sf = g_new0(struct sshfs_file, 1);
     list_init(&sf->write_reqs);
     pthread_cond_init(&sf->write_finished, NULL);
     /* Assume random read after open */
     sf->is_seq = 0;
     sf->next_pos = 0;
+    sf->connver = connver;
     buf_init(&buf, 0);
     buf_add_path(&buf, path);
     buf_add_uint32(&buf, pflags);
-    buf_add_uint32(&buf, 0);
+    buf_add_uint32(&buf, SSH_FILEXFER_ATTR_PERMISSIONS);
+    buf_add_uint32(&buf, mode);
     err = sftp_request(SSH_FXP_OPEN, &buf, SSH_FXP_HANDLE, &sf->handle);
     if (!err) {
         buf_finish(&sf->handle);
@@ -1155,6 +1459,11 @@ static int sshfs_open(const char *path, struct fuse_file_info *fi)
     return err;
 }
 
+static int sshfs_open(const char *path, struct fuse_file_info *fi)
+{
+    return sshfs_open_common(path, 0, fi);
+}
+
 static int sshfs_flush(const char *path, struct fuse_file_info *fi)
 {
     int err;
@@ -1162,6 +1471,9 @@ static int sshfs_flush(const char *path, struct fuse_file_info *fi)
     struct list_head write_reqs;
     struct list_head *curr_list;
 
+    if (!sshfs_file_is_conn(sf))
+        return -EIO;
+
     if (sync_write)
         return 0;
 
@@ -1192,8 +1504,10 @@ static int sshfs_release(const char *path, struct fuse_file_info *fi)
 {
     struct sshfs_file *sf = (struct sshfs_file *) fi->fh;
     struct buffer *handle = &sf->handle;
-    sshfs_flush(path, fi);
-    sftp_request(SSH_FXP_CLOSE, handle, 0, NULL);
+    if (sshfs_file_is_conn(sf)) {
+        sshfs_flush(path, fi);
+        sftp_request(SSH_FXP_CLOSE, handle, 0, NULL);
+    }
     buf_free(handle);
     chunk_put_locked(sf->readahead);
     g_free(sf);
@@ -1233,7 +1547,9 @@ static int sshfs_sync_read(struct sshfs_file *sf, char *rbuf, size_t size,
 static void sshfs_read_end(struct request *req)
 {
     struct read_chunk *chunk = (struct read_chunk *) req->data;
-    if (req->replied) {
+    if (req->error)
+        chunk->res = req->error;
+    else if (req->replied) {
         chunk->res = -EPROTO;
 
         if (req->reply_type == SSH_FXP_STATUS) {
@@ -1259,14 +1575,13 @@ static void sshfs_read_end(struct request *req)
         chunk->res = -EIO;
 
     sem_post(&chunk->ready);
-    chunk_put_locked(chunk);
+    chunk_put(chunk);
 }
+
 static void sshfs_read_begin(struct request *req)
 {
     struct read_chunk *chunk = (struct read_chunk *) req->data;
-    pthread_mutex_lock(&lock);
     chunk->refs++;
-    pthread_mutex_unlock(&lock);
 }
 
 static int sshfs_send_async_read(struct sshfs_file *sf,
@@ -1390,6 +1705,10 @@ static int sshfs_read(const char *path, char *rbuf, size_t size, off_t offset,
 {
     struct sshfs_file *sf = (struct sshfs_file *) fi->fh;
     (void) path;
+
+    if (!sshfs_file_is_conn(sf))
+        return -EIO;
+
     if (sync_read)
         return sshfs_sync_read(sf, rbuf, size, offset);
     else
@@ -1399,9 +1718,7 @@ static int sshfs_read(const char *path, char *rbuf, size_t size, off_t offset,
 static void sshfs_write_begin(struct request *req)
 {
     struct sshfs_file *sf = (struct sshfs_file *) req->data;
-    pthread_mutex_lock(&lock);
     list_add(&req->list, &sf->write_reqs);
-    pthread_mutex_unlock(&lock);
 }
 
 static void sshfs_write_end(struct request *req)
@@ -1409,8 +1726,9 @@ static void sshfs_write_end(struct request *req)
     uint32_t serr;
     struct sshfs_file *sf = (struct sshfs_file *) req->data;
 
-    pthread_mutex_lock(&lock);
-    if (req->replied) {
+    if (req->error)
+        sf->write_error = req->error;
+    else if (req->replied) {
         if (req->reply_type != SSH_FXP_STATUS)
             fprintf(stderr, "protocol error\n");
         else if (buf_get_uint32(&req->reply, &serr) != -1 && serr != SSH_FX_OK)
@@ -1418,7 +1736,6 @@ static void sshfs_write_end(struct request *req)
     }
     list_del(&req->list);
     pthread_cond_broadcast(&sf->write_finished);
-    pthread_mutex_unlock(&lock);
 }
 
 static int sshfs_write(const char *path, const char *wbuf, size_t size,
@@ -1431,6 +1748,10 @@ static int sshfs_write(const char *path, const char *wbuf, size_t size,
     struct buffer *handle = &sf->handle;
 
     (void) path;
+
+    if (!sshfs_file_is_conn(sf))
+        return -EIO;
+
     data.p = (uint8_t *) wbuf;
     data.len = size;
     buf_init(&buf, 0);
@@ -1446,36 +1767,75 @@ static int sshfs_write(const char *path, const char *wbuf, size_t size,
     return err ? err : (int) size;
 }
 
-static int sftp_init()
+static int sshfs_statfs(const char *path, struct statfs *buf)
 {
-    int res = -1;
-    uint8_t type;
-    uint32_t version;
-    struct buffer buf;
-    buf_init(&buf, 4);
-    buf_add_uint32(&buf, PROTO_VERSION);
-    if (sftp_send(SSH_FXP_INIT, &buf) == -1)
-        goto out;
-    buf_clear(&buf);
-    if (sftp_read(&type, &buf) == -1)
-        goto out;
-    if (type != SSH_FXP_VERSION) {
-        fprintf(stderr, "protocol error\n");
-        goto out;
-    }
-    if (buf_get_uint32(&buf, &version) == -1)
-        goto out;
-    if (version != PROTO_VERSION) {
-        fprintf(stderr, "server version: %i, we need: %i\n",
-                version, PROTO_VERSION);
-        goto out;
-    }
-    res = 0;
- out:
-    buf_free(&buf);
-    return res;
+    (void) path;
+
+    buf->f_namelen = 255;
+    buf->f_bsize = 512;
+    buf->f_blocks = 999999999 * 2;
+    buf->f_bfree =  999999999 * 2;
+    buf->f_bavail = 999999999 * 2;
+    buf->f_files =  999999999;
+    buf->f_ffree =  999999999;
+    return 0;
 }
 
+#if FUSE_VERSION >= 25
+static int sshfs_create(const char *path, mode_t mode,
+                        struct fuse_file_info *fi)
+{
+    return sshfs_open_common(path, mode, fi);
+}
+
+static int sshfs_ftruncate(const char *path, off_t size,
+                           struct fuse_file_info *fi)
+{
+    int err;
+    struct buffer buf;
+    struct sshfs_file *sf = (struct sshfs_file *) fi->fh;
+
+    (void) path;
+
+    if (!sshfs_file_is_conn(sf))
+        return -EIO;
+
+    buf_init(&buf, 0);
+    buf_add_buf(&buf, &sf->handle);
+    buf_add_uint32(&buf, SSH_FILEXFER_ATTR_SIZE);
+    buf_add_uint64(&buf, size);
+    err = sftp_request(SSH_FXP_FSETSTAT, &buf, SSH_FXP_STATUS, NULL);
+    buf_free(&buf);
+
+    return err;
+}
+
+static int sshfs_fgetattr(const char *path, struct stat *stbuf,
+                           struct fuse_file_info *fi)
+{
+    int err;
+    struct buffer buf;
+    struct buffer outbuf;
+    struct sshfs_file *sf = (struct sshfs_file *) fi->fh;
+
+    (void) path;
+
+    if (!sshfs_file_is_conn(sf))
+        return -EIO;
+
+    buf_init(&buf, 0);
+    buf_add_buf(&buf, &sf->handle);
+    err = sftp_request(SSH_FXP_FSTAT, &buf, SSH_FXP_ATTRS, &outbuf);
+    if (!err) {
+        if (buf_get_attrs(&outbuf, stbuf, NULL) == -1)
+            err = -EPROTO;
+        buf_free(&outbuf);
+    }
+    buf_free(&buf);
+    return err;
+}
+#endif
+
 static int processing_init(void)
 {
     pthread_mutex_init(&lock, NULL);
@@ -1489,6 +1849,9 @@ static int processing_init(void)
 
 static struct fuse_cache_operations sshfs_oper = {
     .oper = {
+#ifdef SSHFS_USE_INIT
+        .init       = sshfs_init,
+#endif
         .getattr    = sshfs_getattr,
         .readlink   = sshfs_readlink,
         .mknod      = sshfs_mknod,
@@ -1507,6 +1870,12 @@ static struct fuse_cache_operations sshfs_oper = {
         .release    = sshfs_release,
         .read       = sshfs_read,
         .write      = sshfs_write,
+        .statfs     = sshfs_statfs,
+#if FUSE_VERSION >= 25
+        .create     = sshfs_create,
+        .ftruncate  = sshfs_ftruncate,
+        .fgetattr   = sshfs_fgetattr,
+#endif
     },
     .cache_getdir = sshfs_getdir,
 };
@@ -1522,13 +1891,24 @@ static void usage(const char *progname)
 "    -V                     show version information\n"
 "    -p PORT                equivalent to '-o port=PORT'\n"
 "    -C                     equivalent to '-o compression=yes'\n"
+"    -1                     equivalent to '-o ssh_protocol=1'\n"
+"    -o reconnect           reconnect to server\n"
 "    -o sshfs_sync          synchronous writes\n"
 "    -o no_readahead        synchronous reads (no speculative readahead)\n"
 "    -o sshfs_debug         print some debugging information\n"
 "    -o cache=YESNO         enable caching {yes,no} (default: yes)\n"
 "    -o cache_timeout=N     sets timeout for caches in seconds (default: 20)\n"
 "    -o cache_X_timeout=N   sets timeout for {stat,dir,link} cache\n"
+"    -o workaround=LIST     colon separated list of workarounds\n"
+"             none             no workarounds enabled (default)\n"
+"             all              all workarounds enabled\n"
+"             rename           work around problem renaming to existing file\n"
+"    -o idmap=TYPE          user/group ID mapping, possible types are:\n"
+"             none             no translation of the ID space (default)\n"
+"             user             only translate UID of connecting user\n"
 "    -o ssh_command=CMD     execute CMD instead of 'ssh'\n"
+"    -o ssh_protocol=N      ssh protocol to use (default: 2)\n"
+"    -o sftp_server=SERV    path to sftp server or subsystem (default: sftp)\n"
 "    -o directport=PORT     directly connect to PORT bypassing ssh\n"
 "    -o SSHOPT=VAL          ssh options (see man ssh_config)\n"
 "\n", progname);
@@ -1538,7 +1918,6 @@ static void usage(const char *progname)
 
 int main(int argc, char *argv[])
 {
-    char *host = NULL;
     char *fsname;
     int res;
     int argctr;
@@ -1574,6 +1953,11 @@ int main(int argc, char *argv[])
                     arg = "-oCompression=yes";
                 break;
 
+            case '1':
+                if (!arg[2])
+                    arg = "-ossh_protocol=1";
+                break;
+
             case 'p':
                 if (arg[2] || argctr + 1 < argc) {
                     char *newarg;
@@ -1614,6 +1998,42 @@ int main(int argc, char *argv[])
         sync_read = 1;
     if (sshfs_opts[SOPT_DEBUG].present)
         debug = 1;
+    if (sshfs_opts[SOPT_WORKAROUND].present) {
+        struct opt *o = &sshfs_opts[SOPT_WORKAROUND];
+        char *s = opt_get_string(o);
+        if (!s)
+            exit(1);
+
+        for (s = strtok(s, ":"); s; s = strtok(NULL, ":")) {
+            if (strcmp(s, "none") == 0)
+                ;
+            else if (strcmp(s, "all") == 0)
+                rename_workaround = 1;
+            else if (strcmp(s, "rename") == 0)
+                rename_workaround = 1;
+            else {
+                fprintf(stderr, "Invalid value for '%s' option\n", o->optname);
+                exit(1);
+            }
+        }
+    }
+    if (sshfs_opts[SOPT_IDMAP].present) {
+        struct opt *o = &sshfs_opts[SOPT_IDMAP];
+        char *idmap = opt_get_string(o);
+        if (!idmap)
+            exit(1);
+
+        if (strcmp(idmap, "none") == 0)
+            detect_uid = 0;
+        else if (strcmp(idmap, "user") == 0)
+            detect_uid = 1;
+        else {
+            fprintf(stderr, "Invalid value for '%s' option\n", o->optname);
+            exit(1);
+        }
+    }
+    if (sshfs_opts[SOPT_RECONNECT].present)
+        reconnect = 1;
     if (sshfs_opts[SOPT_MAX_READ].present) {
         unsigned val;
         if (opt_get_unsigned(&sshfs_opts[SOPT_MAX_READ], &val) == -1)
@@ -1622,6 +2042,15 @@ int main(int argc, char *argv[])
         if (val < max_read)
             max_read = val;
     }
+    if (sshfs_opts[SOPT_PROTOCOL].present) {
+        if (opt_get_unsigned(&sshfs_opts[SOPT_PROTOCOL], &ssh_ver) == -1)
+            exit(1);
+    }
+    if (sshfs_opts[SOPT_SFTPSERVER].present) {
+        sftp_server = opt_get_string(&sshfs_opts[SOPT_SFTPSERVER]);
+        if (!sftp_server)
+            exit(1);
+    }
     if (sshfs_opts[SOPT_DIRECTPORT].present)
         res = connect_to(host, sshfs_opts[SOPT_DIRECTPORT].value);
     else {
@@ -1631,11 +2060,15 @@ int main(int argc, char *argv[])
     if (res == -1)
         exit(1);
 
-    g_free(host);
     res = sftp_init();
     if (res == -1)
         exit(1);
 
+#ifndef SSHFS_USE_INIT
+    if (detect_uid)
+        sftp_detect_uid();
+#endif
+
     res = processing_init();
     if (res == -1)
         exit(1);
@@ -1644,6 +2077,8 @@ int main(int argc, char *argv[])
     if (res == -1)
         exit(1);
 
+    randseed = time(0);
+
     newargv[newargc++] = g_strdup_printf("-omax_read=%u", max_read);
     newargv[newargc++] = g_strdup_printf("-ofsname=sshfs#%s", fsname);
     g_free(fsname);
